#include <fstream>
#include <algorithm>
#include <ReTexture/rTexture.h>

#define STB_IMAGE_IMPLEMENTATION
#include <stb_image.h>

void RTexture::load(const std::string& file) {
    std::ifstream ifStream(file, std::ios::in | std::ios::binary);
    unsigned char bmpFileHeader[14];

    ifStream.read(reinterpret_cast<char*>(bmpFileHeader), 14);
    if (bmpFileHeader[0] == 'B' && bmpFileHeader[1] == 'M') {
        ifStream.close();
        loadBMP(file);
        return;
    }

    //TODO
    if (file.ends_with(".png"))
        loadPNG(file);
}

void RTexture::loadBMP(const std::string& file) {
    std::ifstream bmpFile(file, std::ios::in | std::ios::binary);
    if (!bmpFile.is_open())
        return;

    unsigned char bmpFileHeader[14];
    unsigned char bmpInfoHeader[40];

    bmpFile.read(reinterpret_cast<char*>(bmpFileHeader), 14);
    bmpFile.read(reinterpret_cast<char*>(bmpInfoHeader), 40);

    width = bmpInfoHeader[4] + (bmpInfoHeader[5] << 8) + (bmpInfoHeader[6] << 16) + (bmpInfoHeader[7] << 24);
    height = bmpInfoHeader[8] + (bmpInfoHeader[9] << 8) + (bmpInfoHeader[10] << 16) + (bmpInfoHeader[11] << 24);

    int rowPadded = (width * 3 + 3) & (~3);
    pixelData.resize(width * height * 3);

    std::vector<unsigned char> rowData(rowPadded);
    for (int y = height - 1; y >= 0; --y) {
        bmpFile.read(reinterpret_cast<char*>(rowData.data()), rowPadded);
        for (int x = 0; x < width; ++x) {
            pixelData[(y * width + x) * 3 + 2] = rowData[x * 3 + 0];
            pixelData[(y * width + x) * 3 + 1] = rowData[x * 3 + 1];
            pixelData[(y * width + x) * 3 + 0] = rowData[x * 3 + 2];
        }
    }
    bmpFile.close();
    format = RTextureFormat::RGB;
}

void RTexture::loadPNG(const std::string &file) {
    int channels, w, h;

    unsigned char *imageData = stbi_load(file.c_str(), &w, &h, &channels, 0);

    if (imageData == nullptr)
        return;
    
    width = w;
    height = h;
    if (channels == 3)
        format = RTextureFormat::RGB;

    if (channels == 4)
        format = RTextureFormat::RGBA;

    pixelData.assign(imageData, imageData + width * height * channels);
    stbi_image_free(imageData);
}

RTexture::RTexture(const std::string &file, const std::vector<RTextureFlag> &args) {
    load(file);
    if (std::find(args.begin(), args.end(), RTextureFlag::INVERT_Y) != args.end())
        invertY();
}

void RTexture::invertY() {
    unsigned int rowSize;
    if (format == RTextureFormat::RGB)
        rowSize = width * 3;
    if (format == RTextureFormat::RGBA)
        rowSize = width * 4;

    std::vector<unsigned char> temp(rowSize);
    for (uint y = 0; y < height / 2; ++y) {
        unsigned char* topRow = &pixelData[y * rowSize];
        unsigned char* bottomRow = &pixelData[(height - y - 1) * rowSize];
        std::copy(bottomRow, bottomRow + rowSize, temp.begin());
        std::copy(topRow, topRow + rowSize, bottomRow);
        std::copy(temp.begin(), temp.end(), topRow);
    }
}

RTexture::RTexture(const std::string& file) {
    *this = RTexture(file, {});
}
